@@ -91,6 +91,82 @@ Each of these can be enabled or disabled later by calling
 		 (const :tag "Both" t))
   :group 'sage-view)
 
+(defvar sage-view-inline-plots-enabled nil)
+(defvar sage-view-inline-output-enabled nil)
+(defcustom sage-view-inline-plots-method 'embedded-mode
+  "Determine how plots are included inline.
+
+This has an effect when enabling and disabling inline plots and
+changing it's value before disabling will disable the new method,
+leaving the old method in place.
+
+If equal to the symbol `emacsclient' then it will use
+emacsclient (as specified by `sage-view-emacsclient') to open the
+file.  This should be robust across Sage versions, but requires
+`sage-view-emacsclient' to be set correctly in order to work.
+
+If equal to the symbol `emacsclient-window' then it will use
+emacsclient as above, but open in a different window rather than
+inline with the rest of the output.
+
+If equal to the symbol `doctest-mode' then it will use
+DOCTEST_MODE.  This is known to work on older versions of Sage,
+but has problems in recent versions.
+
+If equal to the symbol `embedded-mode' then it will use
+EMBEDDED_MODE.  This is known to cause problems with some Sage
+functionality like trace, since Sage thinks it's running in a
+browser."
+  :type '(choice (const :tag "Emacs client inline" emacsclient)
+		 (const :tag "Emacs client in separate window" emacsclient-window)
+		 (const :tag "Doctest mode" doctest-mode)
+		 (const :tag "Embedded mode" embedded-mode))
+  :set (lambda (symbol value)
+	 (if (not (buffer-live-p sage-buffer))
+	     (set symbol value)
+	   (with-current-buffer sage-buffer
+	     (let ((enabled sage-view-inline-plots-enabled))
+	       (when enabled
+		 (sage-view-disable-inline-plots))
+	       (set symbol value)
+	       (when enabled
+		 (sage-view-enable-inline-plots))))))
+  :group 'sage-view)
+
+(defcustom sage-view-emacsclient
+  ;; Magit goes to quite some length to find the right executable, so
+  ;; use that if we have it.
+  (or (and (fboundp 'magit-locate-emacsclient)
+	   (magit-locate-emacsclient))
+      (and invocation-directory
+	   (file-executable-p (concat invocation-directory "emacsclient"))
+	   (concat invocation-directory "emacsclient"))
+      (and invocation-directory
+	   (file-executable-p (concat invocation-directory "bin/emacsclient"))
+	   (concat invocation-directory "bin/emacsclient"))
+      (executable-find "emacsclient"))
+  "The emacslient executable to use for viewing plots inline.
+
+If `sage-view-inline-plots-method' is set to `emacsclient' or
+`emacsclient-window', then this is used.  If this is set
+incorrectly, then it's known to cause issues, at least on OS X."
+  :type 'string
+  :group 'sage-view)
+
+(defcustom sage-view-emacsclient-wrapper
+  (concat (file-name-directory
+	   (or load-file-name
+	       (buffer-file-name))) "sage-view.sh")
+  "A wrapper script around emacsclient for when `sage-view-inline-plots-method' is emacsclient.
+
+The script calls the function `sage-view-handle-emacslient'
+instead of simply opening the image which is what happens when
+`sage-view-inline-plots-method' is emacsclient.
+
+In most cases you should not have to change this."
+  :type 'string
+  :group 'sage-view)
+
 (defvar sage-view-start-string "<html><\\(?:span class=\"math\"\\|script type=\"math/tex\"\\)>"
   "HTML tags that identify the begining of a math formula in Sage output.")
 
@@ -101,9 +177,11 @@ Each of these can be enabled or disabled later by calling
   "<html>\\(?:<font color='[^']*'>\\)?<img src='cell://\\(sage[0-9]+.png\\)'>\\(?:</font>\\)?</html>"
   "HTML tags that identify a plot in Sage output.")
 
+(defvar sage-view-output-regexp
+  "BEGIN_TEXT:\\(\\(?:.\\|\n\\)*?\\):END_TEXT\nBEGIN_LATEX:\\(\\(?:.\\|\n\\)*?\\):END_LATEX"
+  "Regular expression matching typeset output from BackendEmacs.")
+
 (defvar sage-view-dir-name nil)
-(defvar sage-view-inline-plots-enabled nil)
-(defvar sage-view-inline-output-enabled nil)
 
 (defun sage-view-latex->pdf (ov)
   "Start conversion of the LATEX document associated to OV to PDF.
@@ -262,6 +340,32 @@ See also `sage-view-output-filter'."
         `(lambda (event) (interactive "e")
            (sage-view-context-menu ,ov event)))
       (overlay-put ov 'keymap map)
+      (sage-view-process-overlay ov)))
+  ;; Sage 6.6 and on
+  (goto-char (point-min))
+  (while (re-search-forward sage-view-output-regexp (point-max) t)
+    (let* ((full-beg (match-beginning 0))
+           (full-end (match-end 0))
+	   (text-beg (match-beginning 1))
+           (text-end (match-end 1))
+	   (latex-beg (match-beginning 2))
+           (latex-end (match-end 2))
+	   (text (buffer-substring-no-properties text-beg text-end))
+	   (latex (buffer-substring-no-properties latex-beg latex-end))
+           (ov (make-overlay full-beg full-end
+			     nil nil nil))
+	   (map (make-sparse-keymap)))
+      ;; Delete everything except the text
+      (delete-region text-end full-end)
+      (delete-region full-beg text-beg)
+      ;; Populate the overlay
+      (overlay-put ov 'help-echo "mouse-3: Open contextual menu")
+      (overlay-put ov 'text text)
+      (overlay-put ov 'math latex)
+      (define-key map [mouse-3]
+        `(lambda (event) (interactive "e")
+           (sage-view-context-menu ,ov event)))
+      (overlay-put ov 'keymap map)
       (sage-view-process-overlay ov))))
 
 (defun sage-view-output-filter-process-inline-plots (string)
@@ -309,7 +413,8 @@ Function to be inserted in `comint-output-filter-functions'."
       (when sage-view-inline-plots-enabled
 	(sage-view-output-filter-process-inline-plots string))
       (when (and sage-view-inline-output-enabled
-		 (string-match sage-view-final-string string))
+		 (or (string-match sage-view-final-string string)
+		     (string-match ":END_LATEX" string)))
 	(sage-view-output-filter-process-inline-output string)))))
 
 ;;;###autoload
@@ -322,6 +427,11 @@ when `sage-view' mode is enabled and sage is running."
   (sage-send-command "if hasattr(sys.displayhook, 'set_display'): pretty_print_default(True)" nil t)
   ;; sage 5.12
   (sage-send-command "import IPython.core.ipapi; IPython.core.ipapi.get().magic('display typeset')" nil t)
+  ;; Sage 6.6
+  (sage-send-command "from sage.repl.rich_output.backend_emacs import BackendEmacs" nil t)
+  (sage-send-command "from sage.repl.rich_output import get_display_manager" nil t)
+  (sage-send-command "get_display_manager().switch_backend(BackendEmacs(),shell=get_ipython())" nil t)
+  ;; fallthrough
   ;; Sage 6.2
   (sage-send-command "get_ipython().magic('display typeset')" nil t)
   (setq sage-view-inline-output-enabled t)
@@ -348,7 +458,38 @@ when `sage-view' mode is enabled and sage is running."
 WARNING: this communicates with the sage process.  Only use this
 when `sage-view' mode is enabled and sage is running."
   (interactive)
-  (sage-send-command "sage.plot.plot.EMBEDDED_MODE = True;" nil t)
+  (cond
+   ((or (eq sage-view-inline-plots-method 'emacsclient)
+	(eq sage-view-inline-plots-method 'emacsclient-window))
+    (when (not (server-running-p))
+      (error "Server appears to not be running.  Try adding (server-start) to your .emacs"))
+    ;; Store the old viewer
+    (sage-send-command "_emacs_sage_viewer_ = _emacs_sage_viewer_ if '_emacs_sage_viewer_' in globals() else sage.misc.viewer.viewer.png_viewer()" nil t)
+    (sage-send-command (concat "sage.misc.viewer.viewer.png_viewer('''"
+			       (when (eq sage-view-inline-plots-method 'emacsclient)
+				 sage-view-emacsclient-wrapper)
+			       " "
+			       sage-view-emacsclient
+			       "''')")
+		       nil t))
+   ((eq sage-view-inline-plots-method 'embedded-mode)
+    (sage-send-command "sage.plot.plot.EMBEDDED_MODE = True;" nil t))
+   ((eq sage-view-inline-plots-method 'doctest-mode)
+    (sage-send-command "sage.plot.plot.DOCTEST_MODE = True;" nil t)
+    ;; sage 4.7
+    (sage-send-command
+     (format "sage.plot.plot.DOCTEST_MODE_FILE = '%s/sage-view.png';"
+	     sage-view-dir-name)
+     nil t)
+    ;; sage 5.0
+    (sage-send-command
+     (format "if hasattr(sage.plot,'graphics'): sage.plot.graphics.DOCTEST_MODE_FILE = '%s/sage-view.png';\n"
+	     sage-view-dir-name)
+     nil t)
+    ;; sage 5.something
+    (sage-send-command "if hasattr(sage,'doctest'): sage.doctest.DOCTEST_MODE = True;\n" nil t))
+   (t
+    (error "Unknown value for `sage-view-inline-plots-method'")))
   (setq sage-view-inline-plots-enabled t)
   (sage-view-update-modeline))
 
@@ -358,16 +499,48 @@ when `sage-view' mode is enabled and sage is running."
 WARNING: this communicates with the sage process.  Only use this
 when `sage-view' mode is enabled and sage is running."
   (interactive)
-  (sage-send-command "sage.plot.plot.DOCTEST_MODE = False;" nil t)
-  ;; sage 4.7
-  (sage-send-command "sage.plot.plot.DOCTEST_MODE_FILE = None;" nil t)
-  ;; sage 5.0
-  (sage-send-command "if hasattr(sage.plot,'graphics'): sage.plot.graphics.DOCTEST_MODE_FILE = None;\n" nil t)
-  ;; sage 5.something
-  (sage-send-command "if hasattr(sage,'doctest'): sage.doctest.DOCTEST_MODE = False;\n" nil t)
+  (cond
+   ((or (eq sage-view-inline-plots-method 'emacsclient)
+	(eq sage-view-inline-plots-method 'emacsclient-window))
+    (sage-send-command (concat "sage.misc.viewer.viewer.png_viewer(_emacs_sage_viewer_)")
+		       nil t))
+   ((eq sage-view-inline-plots-method 'embedded-mode)
+    (sage-send-command "sage.plot.plot.EMBEDDED_MODE = False;" nil t))
+   ((eq sage-view-inline-plots-method 'doctest-mode)
+    (sage-send-command "sage.plot.plot.DOCTEST_MODE = False;" nil t)
+    ;; sage 4.7
+    (sage-send-command "sage.plot.plot.DOCTEST_MODE_FILE = None;" nil t)
+    ;; sage 5.0
+    (sage-send-command "if hasattr(sage.plot,'graphics'): sage.plot.graphics.DOCTEST_MODE_FILE = None;\n" nil t)
+    ;; sage 5.something
+    (sage-send-command "if hasattr(sage,'doctest'): sage.doctest.DOCTEST_MODE = False;\n" nil t))
+   (t
+    (error "Unknown value for `sage-view-inline-plots-method'")))
   (setq sage-view-inline-plots-enabled nil)
   (sage-view-update-modeline))
 
+(defun sage-view-handle-emacslient (image-file)
+  (when (buffer-live-p sage-buffer)
+    (with-current-buffer sage-buffer
+      (save-excursion
+	(goto-char comint-last-output-start)
+	;; Go to a blank line (where it should be output, but if not don't worry)
+	(re-search-backward "^$" comint-last-input-end t)
+	(let* ((inhibit-read-only t)
+	       (im (create-image image-file 'png))
+	       (p (point))
+	       (junk (insert (concat "image:" image-file)))
+	       (ov (make-overlay p (point) nil nil nil))
+	       (map (make-sparse-keymap)))
+	  (overlay-put ov 'display im)
+	  ;; help alignment as much as possible
+	  (overlay-put ov 'before-string "\n")
+	  (overlay-put ov 'after-string "\n")
+	  (define-key map [mouse-3]
+	    `(lambda (event) (interactive "e")
+	       (sage-view-plot-context-menu ,ov event)))
+	  (overlay-put ov 'keymap map))))))
+
 (defun sage-view-create-temp ()
   "Create a temporary directory and set `sage-view-dir-name'
 to its name.
@@ -411,7 +584,14 @@ writable directory."
    "" str))
 
 (defun sage-view-copy-text (ov)
-  "Copy LATEX source of OV into the kill buffer."
+  "Copy text source of OV into the kill buffer."
+  (let ((text (overlay-get ov 'text)))
+    (if text
+	(kill-new text)
+      (message "No text available"))))
+
+(defun sage-view-copy-latex (ov)
+  "Copy LaTeX source of OV into the kill buffer."
   (let ((text (overlay-get ov 'math)))
     (if text
 	(kill-new (sage-view-cleanup-copied-text text))
@@ -464,7 +644,8 @@ Make sure that there is a valid image associated with OV with
   (popup-menu
    `("Sage View Mode"
      ["Regenerate" (lambda () (interactive) (sage-view-regenerate ,ov))]
-     ["Copy LaTeX" (lambda () (interactive) (sage-view-copy-text ,ov))]
+     ["Copy Text"  (lambda () (interactive) (sage-view-copy-text ,ov))]
+     ["Copy LaTeX" (lambda () (interactive) (sage-view-copy-latex ,ov))]
      ["Save As..." (lambda () (interactive) (sage-view-save-image ,ov))
       `(sage-view-overlay-activep ,ov)]
      ["Zoom in" (lambda (multiplier) (interactive "p")
